/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ble_sdk_uart_over_ble_main main.c
 * @{
 * @ingroup  ble_sdk_app_nus_eval
 * @brief    UART over BLE application main file.
 *
 * This file contains the source code for a sample application that uses the Nordic UART service.
 * This application uses the @ref srvlib_conn_params module.
 */

#include <stdint.h>
#include <string.h>
#include "nrf.h"
#include "ble_hci.h"
#include "ble_advdata.h"
#include "ble_nus.h"
#include "app_util_platform.h"
#include "boards.h"
#include "nrf_sdm.h"
#include "nrf_log_ctrl.h"
#include "nrf_log.h"

#define IS_SRVC_CHANGED_CHARACT_PRESENT 0                                           /**< Include the service_changed characteristic. If not enabled, the server's database cannot be changed for the lifetime of the device. */

#define APP_FEATURE_NOT_SUPPORTED       BLE_GATT_STATUS_ATTERR_APP_BEGIN + 2        /**< Reply when unsupported features are requested. */

#define CENTRAL_LINK_COUNT              0                                           /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT           2                                           /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

#define DEVICE_NAME                     "Dual Peri"                               /**< Name of device. Will be included in the advertising data. */
#define NUS_SERVICE_UUID_TYPE           BLE_UUID_TYPE_VENDOR_BEGIN                  /**< UUID type for the Nordic UART Service (vendor specific). */

#define APP_ADV_INTERVAL                64                                          /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      180                                         /**< The advertising timeout (in units of seconds). */

#define APP_TIMER_PRESCALER             0                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE         4                                           /**< Size of timer operation queues. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(20, UNIT_1_25_MS)             /**< Minimum acceptable connection interval (20 ms), Connection interval uses 1.25 ms units. */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(75, UNIT_1_25_MS)             /**< Maximum acceptable connection interval (75 ms), Connection interval uses 1.25 ms units. */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds), Supervision Timeout uses 10 ms units. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

#define BLE_STACK_EVT_MSG_BUF_SIZE       (sizeof(ble_evt_t) + (23))     /**< Size of BLE event message buffer. This will be provided to the SoftDevice while fetching an event. */

static ble_nus_t                        m_nus;                                      /**< Structure to identify the Nordic UART Service. */

static uint16_t                         m_conn_handle_peripheral_0 = BLE_CONN_HANDLE_INVALID, m_conn_handle_peripheral_1 = BLE_CONN_HANDLE_INVALID;

static ble_uuid_t                       m_adv_uuids[] = {{BLE_UUID_NUS_SERVICE, NUS_SERVICE_UUID_TYPE}};  /**< Universally unique service identifier. */

static uint32_t BLE_EVT_BUFFER[CEIL_DIV(BLE_STACK_EVT_MSG_BUF_SIZE, sizeof(uint32_t))];

static uint16_t m_ble_evt_buffer_size = sizeof(BLE_EVT_BUFFER);


/**@brief Function for assert macro callback.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyse
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num    Line number of the failing ASSERT call.
 * @param[in] p_file_name File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


/**@brief Function for the GAP initialization.
 *
 * @details This function will set up all the necessary GAP (Generic Access Profile) parameters of
 *          the device. It also sets the permissions and appearance.
 */
static void gap_params_init(void)
{
    uint32_t                err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *) DEVICE_NAME,
                                          strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling the data from the Nordic UART Service.
 *
 * @details This function will process the data received from the Nordic UART BLE Service and send
 *          it to the UART module.
 *
 * @param[in] p_nus    Nordic UART Service structure.
 * @param[in] p_data   Data to be send to UART module.
 * @param[in] length   Length of the data.
 */
/**@snippet [Handling the data received over BLE] */
static void nus_data_handler(ble_nus_t * p_nus, uint8_t * p_data, uint16_t length)
{
    uint32_t err_code;
    
    if (p_nus->conn_handle == m_conn_handle_peripheral_0)
    {
        NRF_LOG_INFO("Received following data from peripheral 0:\r\n");
        NRF_LOG_HEXDUMP_INFO(p_data, length);
        if (m_conn_handle_peripheral_1 == BLE_CONN_HANDLE_INVALID)
        {
            NRF_LOG_INFO("Peripheral 1 is not connected:\r\n");
        }
        else
        {
            m_nus.conn_handle = m_conn_handle_peripheral_1;
            err_code = ble_nus_string_send(&m_nus, p_data, length);
            if (err_code == NRF_ERROR_INVALID_STATE)
            {
                NRF_LOG_INFO("Is the CCCD set on peripheral 1?\r\n");
            }
            else
            {
                APP_ERROR_CHECK(err_code);
                NRF_LOG_INFO("The data was sent to peripheral 1\r\n");
            }
        }
    }
    else if (p_nus->conn_handle == m_conn_handle_peripheral_1)
    {
        NRF_LOG_INFO("Received following data from peripheral 1:\r\n");
        NRF_LOG_HEXDUMP_INFO(p_data, length);
        if (m_conn_handle_peripheral_0 == BLE_CONN_HANDLE_INVALID)
        {
            NRF_LOG_INFO("Peripheral 0 is not connected:\r\n");
        }
        else
        {
            m_nus.conn_handle = m_conn_handle_peripheral_0;
            err_code = ble_nus_string_send(&m_nus, p_data, length);
            if (err_code == NRF_ERROR_INVALID_STATE)
            {
                NRF_LOG_INFO("Is the CCCD set on peripheral 0?\r\n");
            }
            else
            {
                APP_ERROR_CHECK(err_code);
                NRF_LOG_INFO("The data was sent to peripheral 0\r\n");
            }
        }
    }
    
}
/**@snippet [Handling the data received over BLE] */


/**@brief Function for initializing services that will be used by the application.
 */
static void services_init(void)
{
    uint32_t       err_code;
    ble_nus_init_t nus_init;

    memset(&nus_init, 0, sizeof(nus_init));

    nus_init.data_handler = nus_data_handler;

    err_code = ble_nus_init(&m_nus, &nus_init);
    APP_ERROR_CHECK(err_code);
}

static uint32_t advertising_start(void)
{
    uint32_t err_code;
    ble_gap_adv_params_t adv_params;
    
    memset(&adv_params, 0, sizeof(adv_params));
    adv_params.interval = APP_ADV_INTERVAL;
    adv_params.timeout  = APP_ADV_TIMEOUT_IN_SECONDS;
    adv_params.type     = BLE_GAP_ADV_TYPE_ADV_IND;

    err_code = sd_ble_gap_adv_start(&adv_params, 0);
    
    return err_code;
}



/**@brief Function for the application's SoftDevice event handler.
 *
 * @param[in] p_ble_evt SoftDevice event.
 */
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t err_code;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            if (m_conn_handle_peripheral_0 == BLE_CONN_HANDLE_INVALID)
            {
                m_conn_handle_peripheral_0 = p_ble_evt->evt.gap_evt.conn_handle;
                NRF_LOG_INFO("Peripheral 0 was connected with connection handle 0x%04X\r\n", m_conn_handle_peripheral_0);
            }
            else if (m_conn_handle_peripheral_1 == BLE_CONN_HANDLE_INVALID)
            {
                m_conn_handle_peripheral_1 = p_ble_evt->evt.gap_evt.conn_handle;
                NRF_LOG_INFO("Peripheral 1 was connected with connection handle 0x%04X\r\n", m_conn_handle_peripheral_1);
            }
            
            if ((m_conn_handle_peripheral_0 == BLE_CONN_HANDLE_INVALID) || (m_conn_handle_peripheral_1 == BLE_CONN_HANDLE_INVALID))
            {
                err_code = advertising_start();
                APP_ERROR_CHECK(err_code);
            }
            break; // BLE_GAP_EVT_CONNECTED

        case BLE_GAP_EVT_DISCONNECTED:
            if (m_conn_handle_peripheral_0 == p_ble_evt->evt.common_evt.conn_handle)
            {
                m_conn_handle_peripheral_0 = BLE_CONN_HANDLE_INVALID;
                NRF_LOG_INFO("Peripheral 0 was disconnected\r\n");
            
            }
            else if (m_conn_handle_peripheral_1 == p_ble_evt->evt.common_evt.conn_handle)
            {
                m_conn_handle_peripheral_1 = BLE_CONN_HANDLE_INVALID;
                NRF_LOG_INFO("Peripheral 1 was disconnected\r\n");
            }
            err_code = advertising_start();
            APP_ERROR_CHECK(err_code);
            break; // BLE_GAP_EVT_DISCONNECTED

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            // Pairing not supported
            err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.common_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
            APP_ERROR_CHECK(err_code);
            break; // BLE_GAP_EVT_SEC_PARAMS_REQUEST

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            // No system attributes have been stored.
            err_code = sd_ble_gatts_sys_attr_set(p_ble_evt->evt.common_evt.conn_handle, NULL, 0, 0);
            APP_ERROR_CHECK(err_code);
            break; // BLE_GATTS_EVT_SYS_ATTR_MISSING

        case BLE_GATTC_EVT_TIMEOUT:
            // Disconnect on GATT Client timeout event.
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break; // BLE_GATTC_EVT_TIMEOUT

        case BLE_GATTS_EVT_TIMEOUT:
            // Disconnect on GATT Server timeout event.
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break; // BLE_GATTS_EVT_TIMEOUT

        case BLE_EVT_USER_MEM_REQUEST:
            err_code = sd_ble_user_mem_reply(p_ble_evt->evt.gattc_evt.conn_handle, NULL);
            APP_ERROR_CHECK(err_code);
            break; // BLE_EVT_USER_MEM_REQUEST

        case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
        {
            ble_gatts_evt_rw_authorize_request_t  req;
            ble_gatts_rw_authorize_reply_params_t auth_reply;

            req = p_ble_evt->evt.gatts_evt.params.authorize_request;

            if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
            {
                if ((req.request.write.op == BLE_GATTS_OP_PREP_WRITE_REQ)     ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_NOW) ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL))
                {
                    if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
                    }
                    else
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
                    }
                    auth_reply.params.write.gatt_status = APP_FEATURE_NOT_SUPPORTED;
                    err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
                                                               &auth_reply);
                    APP_ERROR_CHECK(err_code);
                }
            }
        } break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST
        default:
            // No implementation needed.
            break;
    }
}




/**@brief Function for dispatching a SoftDevice event to all modules with a SoftDevice
 *        event handler.
 *
 * @details This function is called from the SoftDevice event interrupt handler after a
 *          SoftDevice event has been received.
 *
 * @param[in] p_ble_evt  SoftDevice event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    m_nus.conn_handle = p_ble_evt->evt.common_evt.conn_handle;
    ble_nus_on_ble_evt(&m_nus, p_ble_evt);
    on_ble_evt(p_ble_evt);
}


/**@brief Function for the SoftDevice initialization.
 *
 * @details This function initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    uint32_t err_code;
    
    extern uint32_t Image$$RW_IRAM1$$Base;
    const volatile uint32_t ram_start = (uint32_t) &Image$$RW_IRAM1$$Base;
    
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
    err_code = sd_softdevice_enable(&clock_lf_cfg, app_error_fault_handler);
    APP_ERROR_CHECK(err_code);
    
    err_code = sd_nvic_EnableIRQ((IRQn_Type)SD_EVT_IRQn);
    APP_ERROR_CHECK(err_code);
    
    ble_cfg_t ble_cfg;
    ble_gap_cfg_role_count_t ble_gap_cfg_role_count = {PERIPHERAL_LINK_COUNT, 0, 0};
    
    ble_cfg.gap_cfg.role_count_cfg = ble_gap_cfg_role_count;
    
    err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
    APP_ERROR_CHECK(err_code);
    
    uint32_t app_ram_base = ram_start;
    
    err_code = sd_ble_enable(&app_ram_base);
    APP_ERROR_CHECK(err_code);
    
}



/**@brief   Function for handling the Application's BLE Stack events interrupt.
 *
 * @details This function is called whenever an event is ready to be pulled.
 */
void SD_EVT_IRQHandler(void)
{
    bool no_more_ble_evts = false;
    
    for (;;)
    {
        uint32_t err_code;

        // Fetch BLE Events.
        if (!no_more_ble_evts)
        {
            // Pull event from stack
            uint16_t evt_len = m_ble_evt_buffer_size;

            err_code = sd_ble_evt_get((uint8_t *)BLE_EVT_BUFFER, &evt_len);
            if (err_code == NRF_ERROR_NOT_FOUND)
            {
                no_more_ble_evts = true;
            }
            else if (err_code != NRF_SUCCESS)
            {
                APP_ERROR_HANDLER(err_code);
            }
            else
            {
                // Call application's BLE stack event handler.
                ble_evt_dispatch((ble_evt_t *)BLE_EVT_BUFFER);
            }
        }

        if (no_more_ble_evts)
        {
            break;
        }
    }
}

/**@brief Function for initializing the Advertising functionality.
 */
static void advertising_init(void)
{
    uint32_t err_code;
    
    ble_advdata_t ble_advdata;
    ble_advdata_t scan_response;
    
    memset(&ble_advdata, 0, sizeof(ble_advdata));
    ble_advdata.name_type   = BLE_ADVDATA_FULL_NAME;
    ble_advdata.flags       = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;
    
    memset(&scan_response, 0, sizeof(scan_response));
    scan_response.uuids_complete.uuid_cnt = sizeof(m_adv_uuids) / sizeof(m_adv_uuids[0]);
    scan_response.uuids_complete.p_uuids  = m_adv_uuids;

    err_code = ble_advdata_set(&ble_advdata, &scan_response);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for placing the application in low power state while waiting for events.
 */
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}


/**@brief Application main function.
 */
int main(void)
{
    uint32_t err_code;
    
    err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);
    
    NRF_LOG_INFO("Dual Peripheral application!\r\n");

    ble_stack_init();
    gap_params_init();
    services_init();
    advertising_init();
    advertising_start();
    
    // Enter main loop.
    for (;;)
    {
        power_manage();
    }
}


/**
 * @}
 */
